import base58
import cashaddress


class Info:
    TYPE = None  # override
    NAME = None  # override
    data = None  # calculate

    def hexlike(self):
        """Return hex-like string for data as <Payment Type + Payment Data>."""
        return '{:02d}{}'.format(self.TYPE, self.data.lower())

    def __str__(self):
        return '{}\ntype: {}\ndata: {}'.format(self.NAME, self.TYPE, self.data)


class PaymentKeyHash(Info):
    TYPE = 1
    NAME = 'Key Hash'

    def __init__(self, address_string):
        address_string = _loose_address(address_string)
        try:
            address = cashaddress.convert.Address.from_string(address_string)
        except cashaddress.convert.InvalidAddress:
            raise ValueError('unable to interpret address as cashaddress or legacy ({})'
                             ''.format(address_string))
        self.data = _hash160(address)


def _loose_address(address_string):
    if _looks_like_cashaddr_without_prefix(address_string):
        address_string = 'bitcoincash:{}'.format(address_string)
    return address_string


def _looks_like_cashaddr_without_prefix(s):
    if s[0] != 'q':
        return False
    if len(s) != 42:
        return False
    return True


def _hash160(address):
    legacy = address.legacy_address()
    return base58.b58decode_check(legacy).hex()[2:]



import unittest

import cashaccount.payment as pay


class TestKeyHashInfo(unittest.TestCase):
    CASH_ADDR = 'bitcoincash:qrme8l598x49gmjhn92dgwhk5a3znu5wfcf5uf94e9'
    LEGACY_ADDR = '1Pa5CCeYCpWXFJXRnhZpmhJRuFg184HGHz'
    EXPECTED_HASH160 = 'f793fe8539aa546e579954d43af6a76229f28e4e'
    EXPECTED_FULL_INFO = '01' + EXPECTED_HASH160

    def test_works_for_valid_addresses(self):
        self.assertEqual(pay.PaymentKeyHash(self.CASH_ADDR).hexlike(), self.EXPECTED_FULL_INFO)
        self.assertEqual(pay.PaymentKeyHash(self.LEGACY_ADDR).hexlike(), self.EXPECTED_FULL_INFO)

        no_prefix = self.CASH_ADDR.replace('bitcoincash:', '')
        self.assertEqual(pay.PaymentKeyHash(no_prefix).hexlike(), self.EXPECTED_FULL_INFO)

    def test_raises_some_exception_for_invalid_addresses(self):
        with self.assertRaises(Exception):
            pay.PaymentKeyHash('invalid address')
        with self.assertRaises(Exception):
            pay.PaymentKeyHash('qstillinvalid')

    def test_string_has_useful_info(self):
        expected = 'Key Hash\n' \
                   'type: 1\n' \
                   'data: {}'.format(self.EXPECTED_HASH160)
        info = pay.PaymentKeyHash(self.CASH_ADDR)
        self.assertEqual(str(info), expected)



import unittest
from unittest import mock

import cashaccount.payment as pay
import cashaccount.registration as rgstr


class TestNameValidation(unittest.TestCase):
    def test_returns_valid_names(self):
        self.assertEqual(rgstr.validate_name('lower'), 'lower')
        self.assertEqual(rgstr.validate_name('UPPER'), 'UPPER')
        self.assertEqual(rgstr.validate_name('mixedCASE'), 'mixedCASE')
        self.assertEqual(rgstr.validate_name('mixedWITH01234'), 'mixedWITH01234')
        self.assertEqual(rgstr.validate_name('mixedWITH__01234'), 'mixedWITH__01234')
        max_length_name = 'a' * 99
        self.assertEqual(rgstr.validate_name(max_length_name), max_length_name)

    def test_raises_ValueError_for_invalid_names(self):
        with self.assertRaises(ValueError):
            rgstr.validate_name('nonalphabet文字')
        with self.assertRaises(ValueError):
            rgstr.validate_name('hyphens-not-allowed')
        with self.assertRaises(ValueError):
            too_long = 'a' * 100
            rgstr.validate_name(too_long)


class TestConverters(unittest.TestCase):
    def test_to_hex(self):
        assert rgstr.to_hexlike('emergent_reasons') == '656d657267656e745f726561736f6e73'

    def test_from_hex(self):
        assert rgstr.from_hexlike('656d657267656e745f726561736f6e73') == 'emergent_reasons'


class TestElectronMarkdown(unittest.TestCase):
    def test_key_hash(self):
        expected = '<push><hex>01010101' \
                   '<push><hex>656d657267656e745f726561736f6e73' \
                   '<push><hex>01f793fe8539aa546e579954d43af6a76229f28e4e'

        name = 'emergent_reasons'
        payment_info = pay.PaymentKeyHash('bitcoincash:qrme8l598x49gmjhn92dgwhk5a3znu5wfcf5uf94e9')
        registration = rgstr.Registration(name, payment_info)
        self.assertEqual(rgstr.electron_markdown(registration), expected)


class TestString(unittest.TestCase):
    def test_has_useful_info(self):
        expected = 'name: emergent_reasons\n' \
                   'payment info:\n' \
                   '  some lines\n' \
                   '  of payment info'
        name = 'emergent_reasons'
        payment_info_mock = mock.MagicMock()
        payment_info_mock.__str__.return_value = 'some lines\nof payment info'
        registration = rgstr.Registration(name, payment_info_mock)
        self.assertEqual(str(registration), expected)



import re
import textwrap


NAME_REGEX = r'^[a-zA-Z0-9_]{1,99}$'


class Registration:
    def __init__(self, name, payment_info):
        self.name = validate_name(name)
        self.payment_info = payment_info

    def __str__(self):
        info = textwrap.indent(str(self.payment_info), '  ')
        return 'name: {}\n' \
               'payment info:\n{}' \
               ''.format(self.name, info)


def validate_name(n):
    if re.match(NAME_REGEX, n):
        return n
    raise ValueError(r'name ({name}) does not meet requirements ({regex})'
                     r''.format(name=n, regex=NAME_REGEX))


def electron_markdown(registration):
    template = '<push><hex>01010101' \
               '<push><hex>{hex_name}' \
               '<push><hex>{payment_info}'
    return template.format(hex_name=to_hexlike(registration.name),
                           payment_info=registration.payment_info.hexlike())


def to_hexlike(s):
    return s.encode('utf-8').hex()


def from_hexlike(hexlike_string):
    return bytes.fromhex(hexlike_string).decode('utf-8')



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = [l for l in f.readlines() if l.strip()]

with open('README.md') as f:
    long_description = f.read()

setup(
    name='pycashaccount',
    version='0.1',
    author='emergent-reasons',
    author_email='emergentreasons@gmail.com',
    description='helper for creating cash accounts',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/emergent-reasons/pycashaccount',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    include_package_data=True,
    install_requires=requirements,
    entry_points='''
        [console_scripts]
        cashaccount=cashaccount.cli:run
    ''',
)



import click

from . import electron_markdown as em
from . import PaymentKeyHash
from . import Registration


@click.group()
def run():
    pass


@run.command()
@click.argument('name')
@click.argument('cash_or_legacy_address')
@click.option('--electron-markdown', is_flag=True, default=False)
def address(name, cash_or_legacy_address, electron_markdown):
    info = PaymentKeyHash(cash_or_legacy_address)
    registration = Registration(name, info)

    if electron_markdown:
        result = em(registration)
    else:
        result = str(registration)
    click.echo(result)
